#
rm(list=ls())

#
setwd("D:/DataMining/Databases_for_mining/dataset_for_soft_dev_and_comparison/discriminant_analysis/R")

#chargement des donnes
library(xlsx)
DTrain <- read.xlsx("Eau_de_vie_LDA_R.xlsx",header = TRUE, sheetIndex = 1)
str(DTrain)

#X
XTrain <- DTrain[-1]
print(str(XTrain))

#y
yTrain <- DTrain$TYPE

#nombre de variables
p <- ncol(XTrain)

#nombre d'observations
n <- nrow(DTrain)

#nombre de classes
K <- nlevels(yTrain)

#moyennes des variables
xb <- colMeans(DTrain[-1])
print(xb)

#lda
library(MASS)
mLda <- lda(TYPE ~ ., data = DTrain)
print(mLda)

#affichage des proprits
print(attributes(mLda))

mLda$scaling
mLda$svd
mLda$means
mLda$prior

#constantes des fonctions canoniques
const_ <- apply(mLda$scaling,2,function(v){-sum(v*xb)})
const_

#avec les fonctions canoniques
#on retrouve les moyennes conditionnelles
cond_Xb <- sapply(1:ncol(mLda$scaling),function(j)
  {sapply(levels(yTrain),
          function(niveau){sum(mLda$scaling[,j]*mLda$means[niveau,])+const_[j]})
  })
colnames(cond_Xb) <- paste("LD",1:ncol(mLda$scaling),sep="")
rownames(cond_Xb) <- levels(DTrain$TYPE)

#affichage
print(cond_Xb)

#coefficients
coef_ <- sapply(levels(yTrain),function(niveau)
  {rowSums(sapply(1:ncol(mLda$scaling),
          function(j){mLda$scaling[,j]*cond_Xb[niveau,j]}))
  })

print(coef_)


#intercept
intercept_ <- sapply(levels(yTrain),function(niveau)
  {sum(mapply(prod,const_,cond_Xb[niveau,]))-0.5*sum(cond_Xb[niveau,]^2)+log(mLda$prior[niveau])})
names(intercept_) <- levels(yTrain)
print(intercept_)

#matrice de convariance totale
TOT <- cov(XTrain)
print(TOT)

#WIT
WIT <- (1.0/(n-K))*Reduce("+",lapply(levels(yTrain),function(niveau){(sum(yTrain==niveau)-1)*cov(XTrain[yTrain==niveau,])}))
print(WIT)

#lambda de Wilks
#matrices intermdiaires
WITprim <- (n-K)/n*WIT 
TOTprim <- (n-1)/n*TOT

LW <- det(WITprim)/det(TOTprim)
print(LW)

#contribution des variables

#ddl numrateur
ddlSuppNum <- K - 1

#ddl dnominateur
ddlSuppDenom <- n - K - p + 1

#vecteur des F
FTest <- numeric(p)

#p-value
pvalueFTest <- numeric(p)

#boucle
for (j in 1:p){
  #Lambda corresp.
  LWvar <- det(WITprim[-j,-j])/det(TOTprim[-j,-j])
  #print(LWvar)
  #F
  FTest[j] <- ddlSuppDenom / ddlSuppNum * (LWvar/LW - 1)
  #p-avlue
  pvalueFTest[j] <- pf(FTest[j],ddlSuppNum,ddlSuppDenom,lower.tail=FALSE)
}

#affichage
temp <- data.frame(var=colnames(XTrain),FValue=FTest,pvalue=round(pvalueFTest,6))
print(temp)
  

#chargement de l'chantillon test
DTest <- read.xlsx("Eau_de_vie_LDA_R.xlsx",header = TRUE, sheetIndex = 2)
str(DTest)

#X
XTest <- DTest[-1]

#y
yTest <- DTest$TYPE

#prediction en test
predTest <- predict(mLda,newdata=XTest)
print(attributes(predTest))

#prdiction brute
print(predTest$class)

#distribution des prdiction
print(table(predTest$class))

#matrice de confusion
mc <- table(yTest,predTest$class)
print(mc)

#taux de reconnaissance (accuracy)
accTest <- sum(diag(mc))/sum(mc)
print(accTest)

#taux d'erreur
errTest <- 1 - accTest
print(errTest)

#rappel (sensibilit par classe)
sensTest <- diag(mc)/rowSums(mc)
print(sensTest)

#prcision par classe
precisionTest <- diag(mc)/colSums(mc)
print(precisionTest)

#on a la mme chose avec caret
library(caret)

#appel de confusion matrix
caret::confusionMatrix(reference=yTest,data=predTest$class)

#coordonnes factorielles
head(predTest$x)

#graphique - mme chelle en abscisse et ordonne

eqscplot(predTest$x[,1],predTest$x[,2],col=c("red","green","blue")[yTest],xlab="LD1",ylab="LD2")
legend("topright",legend=levels(yTest),text.col=c("red","green","blue"))
abline(h=0,col="gray")
abline(v=0,col="gray")

#essayons de les reproduire
#produit matriciel
coordFact <- t(t(mLda$scaling) %*% t(as.matrix(XTest)))

#ajouter les constantes
coordFact <- sapply(1:ncol(mLda$scaling),function(j){coordFact[,j]+const_[j]})

#affichage
head(coordFact)

#proba a posteriori - proba d'affectation
head(predTest$posterior)

#calcul fonction de classement
vfClass <- t(t(coef_) %*% t(as.matrix(XTest)))

#rajouter la constante
vfClass <- sapply(levels(yTrain),function(niveau){vfClass[,niveau]+intercept_[niveau]})
head(vfClass)

#softmax
pClass <- t(apply(vfClass,1,function(ligne){exp(ligne)/sum(exp(ligne))}))
head(pClass)

#vrification individuelle
predTest$posterior[25,]
pClass[25,]

#vrification globale
print(sum(abs(predTest$posterior-pClass)))

#autre piste - distance carre gnralise
#solution un peu naive
# d2G <- t(apply(predTest$x,1,function(ligne){
#   sapply(levels(yTrain),function(niveau){sum((cond_Xb[niveau,]-ligne)^2)-2*log(mLda$prior[niveau])})
#   }))

#autre piste - distance carre gnralise
d2G <- sapply(levels(yTrain),
              function(niveau)
                {rowSums(scale(predTest$x,center=cond_Xb[niveau,],scale=FALSE)^2)-2*log(mLda$prior[niveau])})
head(d2G)

#variante softmax avec la correction -- cf. SAS
pClassD2G <- t(apply(d2G,1,function(ligne){exp(-0.5*ligne)/sum(exp(-0.5*ligne))}))
head(pClassD2G)

#vrification
predTest$posterior[25,]
pClassD2G[25,]

#vrification globale
print(sum(abs(predTest$posterior-pClassD2G)))

library(klaR)
greedy.wilks(TYPE ~ ., data = DTrain, niveau = 0.01)

#un petit code pour la slection backward


#fonction pour le test du retrait d'une variable numro j
#SW : matrice de covariance intra
#ST : matrice de covariance totale
#N : nombre d'observations
#K : nombre de classes
#precLW : valeur du lambda avant le retrait
test.retrait <- function(j,SW,ST,N,K,precLW){
  #nombre de variables  cette tape
  nbVar <- ncol(SW)
  #lamdba
  lambada <- det(SW[-j,-j])/det(ST[-j,-j])
  #ddl
  ddlNum <- K - 1
  ddlDenom <- N - K - nbVar + 1
  #f
  f <- ddlDenom / ddlNum * (lambada / precLW - 1)
  #p-value
  pvalue <- pf(f,ddlNum,ddlDenom,lower.tail=FALSE)
  #renvoyer sous forme de vecteur
  return(c(lambada,f,pvalue))
}


#fonction de slection backward
#DW : matrice initiale de covariance intra
#DT : matrice initiale de covariance totale
#lambdaInit : valeur initiale du lambda de wilks
#alpha : risque pour piloter la slection
# sel.backward <- function(DW,DT,lstInit,lambdaInit,alpha=0.05){
#   #valeurs lambda suite  chaque retrait
#   lambda <- c()
#   #valeur F suite  chaque retrait
#   ftest <- c()
#   #pvalue suite  chaque retrait
#   pvalue <- c()
#   #liste des variables retires
#   lstvarDel <- c()
#   #*******************
#   #boucle de recherche
#   repeat{
#     #resultats d'une tape
#     res <- t(sapply(1:ncol(DW),retrait,SW=DW,ST=DT,N=n,K=K,precLW=lambdaInit))
#     #trouver la ligne qui maximise la p-value
#     id <- which.max(res[,3])
#     #vrifier si retrait
#     if (res[id,3] > alpha){
#       #nom de la variable  retirer
#       lstvarDel <- c(lstvarDel,lstInit[id])
#       #lambda suite au retrait
#       lambda <- c(lambda,res[id,1])
#       #F corresp. au test
#       ftest <- c(ftest,res[id,2])
#       #p-value
#       pvalue <- c(pvalue,res[id,3])
#       #retirer
#       lstInit <- lstInit[-id]
#       #si plus de variables
#       if (length(lstInit) == 0){
#         #arrt puisque plus de variables  retirer
#         break
#       } else {
#         #retirer les colonnes des matrices
#         DW <- DW[-id,-id]
#         DT <- DT[-id,-id]
#         #mj du lambda
#         lambdaInit <- res[id,1]
#       }
#     } else {
#       #plus de retrait
#       break
#     }
#   }
#  #resultats sous la forme d'un data frame
#  resultats <- data.frame(var=lstvarDel,L=lambda,f=ftest,pval=pvalue)
#  return(resultats) 
# }

#fonction de slection backward
#DW : matrice initiale de covariance intra
#DT : matrice initiale de covariance totale
#lambdaInit : valeur initiale du lambda de wilks
#alpha : risque pour piloter la slection
sel.backward <- function(DW,DT,lstInit,lambdaInit,alpha=0.05){
  #matrice pour rcuprer les rsultats  chaque tape
  mResult <- matrix(0,nrow=0,ncol=3)
  colnames(mResult) <- c('Lambda','F-Test','p-value')
  #liste des variables retires
  lstvarDel <- c()
  #*******************
  #boucle de recherche
  repeat{
    #resultats d'une tape
    res <- t(sapply(1:ncol(DW),test.retrait,SW=DW,ST=DT,N=n,K=K,precLW=lambdaInit))
    #trouver la ligne qui maximise la p-value
    id <- which.max(res[,3])
    #vrifier si retrait (comparer pvalue  alpha)
    if (res[id,3] > alpha){
      #nom de la variable  retirer
      lstvarDel <- c(lstvarDel,lstInit[id])
      #rajouter la ligne de rsultats (lambda,F-Test,p-value)
      mResult <- rbind(mResult,res[id,])
      #retirer
      lstInit <- lstInit[-id]
      #si plus de variables
      if (length(lstInit) == 0){
        #arrt puisque plus de variables  retirer
        break
      } else {
        #retirer les colonnes des matrices
        DW <- DW[-id,-id]
        DT <- DT[-id,-id]
        #mj du lambda
        lambdaInit <- res[id,1]
      }
    } else {
      #plus de retrait parce que condition alpha
      break
    }
  }
  #resultats sous la forme d'un data frame
  resultats <- data.frame(var=lstvarDel,mResult)
  return(resultats) 
}

#lancer la slection
removedVar <- sel.backward(WITprim,TOTprim,colnames(XTrain),LW,alpha=0.01)
print(removedVar)

#variables restantes
resteVar <- setdiff(colnames(XTrain),removedVar$var)
print(resteVar)

#crer une nouvelle formule
formule <- reformulate(resteVar,"TYPE")
print(formule)

#lancer la lda
mLdaBack <- lda(formule,data=DTrain)
print(mLdaBack)
